/*
 *This files contains the Agent declarationa and Timers
 */

//Uncomment to debug 
#define DEBUG

//Uncomment to introduce optimizations
#define REFLECTION


#ifndef __node_num__
#define __node_num__
const int NODENUM=200;
#endif

#ifndef __replica_num__
#define __replica_num__
const int REPLICANUM=10;
#endif

#ifndef __redman_h__
#define __redman_h__

#include "redman_packet.h"
#include "neighbors_table.h"
#include "config.h"
#include "object.h"
#include "agent.h"
#include "timer-handler.h"
#include "address.h"
#include <lib/bsd-list.h>
#include <stdlib.h>
#include "replica.h"
#include "replication_info.h"
#include "message_queue.h"
#include "discharge_queue.h"



class RedmanAgent;

#define NEXT_EXPLORED         1
#define SND_HELLO             2
#define CHECK_NEIGHBORS       3
#define RED_DISC              4
#define RED_ADV               5
#define PROBE_WAIT            6
#define PROBE_REPLY           7
#define SEARCH_PROBE_WAIT     8
#define DELEGATION            9
#define PROBE_NEIGHBORS       10
#define AUM_SEND              11
#define AUM_CHECK             12
#define AUM_EXTRACT           13
#define DISCHARGE_EXTRACT     14

#define DEBUG

/*=========================================================================================================
 *                                             TIMERS
 *=========================================================================================================*/

/*============================================================
 *  BOUNDARIES IDENTIFICATION PROTOCOL
 *============================================================*/

/*
 *The discovery timer sets the time passing between the broadcast of node identities
 *and the moment when nodes broadcast the number of its neighbors
 */

 class REDDiscoveryTimer : public TimerHandler {
 public:
    REDDiscoveryTimer (RedmanAgent* a) :TimerHandler(), a_(a) {};
 protected:
    virtual void expire(Event *);
    RedmanAgent* a_;
 };


/*
 *When the participation timer fire, the node decides whether to participate
 *to the locality or not
 */
class REDPartAdvTimer : public TimerHandler {
 public:
    REDPartAdvTimer (RedmanAgent* a) :TimerHandler(), a_(a) {};
 protected:
    virtual void expire(Event *);
    RedmanAgent* a_;
 };
 

/*============================================================
 *                REPLICA PLACEMENT AND SEARCH
 *============================================================*/

/*
 *The REDMAN Next Explored Timer represents the time wait before the node checks its
 *neighbors characteristics and chooses the most suitable to carry on the exploration
 */

 class NextExploredTimer : public TimerHandler {
 public:
    NextExploredTimer (RedmanAgent* a) :TimerHandler(), a_(a) {};
 protected:
    virtual void expire(Event *);
    RedmanAgent* a_;
 };

/*
 *The REDMAN Degation Timer represents the delay to wait for delegate message delivery
 *(not to collide with the confirmation packets)
 */

 class DisseminationDelegationTimer : public TimerHandler {
 public:
    DisseminationDelegationTimer (RedmanAgent* a) :TimerHandler(), a_(a) {};
 protected:
    virtual void expire(Event *);
    RedmanAgent* a_;
 };
 
/*============================================================
 *                NEIGHBORS PROBE REPLY TIMER
 *============================================================*/

/*
 *The REDMAN Neighbors Probe Wait Timer represents the time wait before the current node
 *compares local neighbors Probe Replies and decides which is the most suitable node
 */

 class ProbeWaitTimer : public TimerHandler {
 public:
    ProbeWaitTimer (RedmanAgent* a) :TimerHandler(), a_(a) {};
 protected:
    virtual void expire(Event *);
    RedmanAgent* a_;
 };


/*
 *The REDMAN Neighbors Probe Reply Timer represents the random interval before the current node
 *sends a Probe Reply to the prober (current candidate replica acceptor)
 */

 class ProbeReplyTimer : public TimerHandler {
 public:
    ProbeReplyTimer (RedmanAgent* a) :TimerHandler(), a_(a) {};
 protected:
    virtual void expire(Event *);
    RedmanAgent* a_;
 };

/*
 *The REDMAN Search Probe Wait Timer represents the time wait before the current node
 *compares local neighbors Probe Replies and decides which is the most suitable node
 */

 class SearchProbeWaitTimer : public TimerHandler {
 public:
    SearchProbeWaitTimer (RedmanAgent* a) :TimerHandler(), a_(a) {};
 protected:
    virtual void expire(Event *);
    RedmanAgent* a_;
 };




/*===============================================================
 *                CURRENT NEIGHBORS DETERMINATION
 *==============================================================*/

/*
 *When this timer fires the node sends its periodical Hello packet
 */
class REDSendHelloTimer : public TimerHandler {
public:
        REDSendHelloTimer(RedmanAgent* a) : a_(a) {};
        void	expire(Event*);
private:
        RedmanAgent    *a_;  
};

/*
 *When this timer fires, the node checks whether it dense MANET
 *participation conditions still hold (i.e. if the number of neighbors
 *exceeds the requested) or not.
 */
class REDCheckNeighborTimer : public TimerHandler {
public:
        REDCheckNeighborTimer(RedmanAgent* a) : a_(a) {}
        void	expire(Event*);
private:
        RedmanAgent    *a_;      
};


/*===============================================================
 *                MANAGER RELATED TIMERS
 *==============================================================*/

/*
 *When this timer fires the node probes its neighbors
 */
class REDDelayedProbeTimer : public TimerHandler {
public:
        REDDelayedProbeTimer(RedmanAgent* a) : a_(a) {};
        void	expire(Event*);
private:
        RedmanAgent    *a_;  
};


/*
 *When this timer fires each delegate sends its shared resource list to the manager
 */
class REDAUMTimer : public TimerHandler {
public:
        REDAUMTimer(RedmanAgent* a) : a_(a) {};
        void	expire(Event*);
private:
        RedmanAgent    *a_;  
};

/*
 *When this timer fires the manager check the resources present in the dense MANET
 */
class REDAUMCheckTimer : public TimerHandler {
public:
        REDAUMCheckTimer(RedmanAgent* a) : a_(a) {};
        void	expire(Event*);
private:
        RedmanAgent    *a_;  
};

/*
 *When this timer fires a single message is extracted from Manager MEssage_Queue after AUM check
 *Messages are sent till the queue is empty
 */
class REDAUMExtractTimer : public TimerHandler {
public:
        REDAUMExtractTimer(RedmanAgent* a) : a_(a) {};
        void	expire(Event*);
private:
        RedmanAgent    *a_;  
};

/*
 *When this timer fires a single message is extracted from Manager MEssage_Queue after AUM check
 *Messages are sent till the queue is empty
 */
class REDDischargeExtractTimer : public TimerHandler {
public:
        REDDischargeExtractTimer(RedmanAgent* a) : a_(a) {};
        void	expire(Event*);
private:
        RedmanAgent    *a_;
};
  


/*======================================================================================================
  NODE
  ======================================================================================================*/
struct node {
        int id;
        };
        



/*=====================================================================================================
 *                                    REDMAN AGENT
 *=====================================================================================================*/
class RedmanAgent : public Agent {

public:

  /*=======================================*
   *           MEMBER FUNCTIONS            *
   *=======================================*/
   
  RedmanAgent(nsaddr_t id);
  int command(int argc, const char*const* argv);
  void recv(Packet*, Handler*);
  void timeout(int);

  //BOUNDARIES IDENTIFICATION PROTOCOL
  void start();
  void recvDensityDiscovery(Packet* pkt);
  void discover();
  void advertise();


  //NEIGHBORS DETECTION PROTOCOL
  void sendHello();
  void recvHello(Packet*);
  void nb_purge();
  
  //REPLICA DISTRIBUTION
  void start_replica_distribution(int resource_id, int num_replicas);  //used by the current delegate
  void send_replica_distribution(int next, int discharge);
  void recv_replica_distribution(Packet* pkt);
  void delegate_replica_distribution();
  //NEIGHBORS PROBE PROTOCOL  
  void recv_probe(Packet* pkt);
  void send_probe_rep();
  void recv_probe_rep(Packet*pkt);
  //CONFIRMATION
  void confirm_neighbors();
  

  //REPLICA SEARCH
  void start_replica_search(int resource_id);
  void send_replica_search(int next);
  void recv_replica_search(Packet* pkt);
  void delegate_search();
  void search_reply(int);
  void recv_replica_search_rep(Packet* pkt);

  int reply_placement(Replica r);
  
  
  
  
  //MANAGER RELATED METHODS
  void recv_degree_req(Packet* pkt);
  void recv_degree_rep(Packet* pkt);
  //The hosts communicate the manager the new replica placed
  void communicate_manager(int resource_id, int isexitingdelegate, int exitingdelegate);
  void probe_neighbors();
  void update_srt(Packet* pkt);
  //The hosts communicate the manager that they are on leave
  void notify_manager();
  void recv_leaving_notification(Packet* pkt);
  //The hosts receive the request for a new replication
  void manager_required_replication(Packet* pkt);
  //NExt methods are related to Asynchronous Update Messages send/recv and check
  void aum_send();
  void aum_check();
  void aum_recv(Packet* pkt);
  
  //General method that extracts the messges after AUM check from a queue and sends
  void extract_from_queue();
  //Discharge messages
  void extract_from_discharge();
  //General method to send a replication request to a node
  void send_replication_request(int, int, int);
  
  
  
  
  

  /*=======================================*
   *           ATTRIBUTES                  *
   *=======================================*/

  /*STATIC*/
  //Intervals
  //Delay to advertise my presence to my neighbors
  static int disc_delay_;
  //Intervals and delays to advertise the number of my neighbors
  static int part_adv_interval_;
  static int part_adv_delay_;
  //Limit of neighbors to participate to the hotspot
  static int limit_;
  //The hello period
  static float hello_period_;
  //The hello delay
  static float hello_delay_;
  //The check period
  static float check_hello_period_;
  //The timer wait before decide which neighbor is mostly suitable in REDMAN Replica Dissemination
  static float probe_wait_interval_;
  //The delay intercurring before a node sends a Probe Reply
  static float probe_reply_delay_;
  //The neighbors limit to reflect
  static int reflection_limit_;
  //The hops limit of the replication
  static int replica_hops_;
  //The hops limit of the searches
  static int search_hops_;
  //The id of the manager
  static int manager_id_;
  //The period of the Asynchronous updates 
  static float aum_period_;
  //The check period of Asynchronous updates
  static float aum_check_period_;
  
  

    

private:
  Packet* BCastPacket();

  //STATE VARIABLES
  //Current node ID
  int nid_;
  //This node has sent the discovery packets
  int discovery_sent_;
  //This node is the initiator
  int initiator_;
  //The number of currently received discovery packets
  int received_discovery_;
  //This node participate in the dense MANET
  int part_;
  
  //TIMERS
  //BOUNDARIES IDENTIFICATION PROTOCOL
  //The discovery timer
  REDDiscoveryTimer disc_timer_;
  //The participation advertisement timer
  REDPartAdvTimer part_adv_timer_;
  //NEIGHBORS MAINTENANCE
  REDSendHelloTimer send_hello_timer_;
  REDCheckNeighborTimer check_neighbor_timer_;
  //Neighbors probe reply timer
  ProbeWaitTimer probe_wait_timer_;
  //Probe Reply timer
  ProbeReplyTimer probe_reply_timer_;
  //Search probe reply timer
  SearchProbeWaitTimer search_probe_wait_timer_;
  //Delegation (delay) timer
  DisseminationDelegationTimer delegation_timer_;
  
  /**MANAGER RELATED TIMERS *****/
  REDDelayedProbeTimer delayed_probe_timer_;
  //AUM Timer
  REDAUMTimer aum_timer_;
  //AUM Check Timer
  REDAUMCheckTimer aum_check_timer_;
  //AUM Message Queue extraction timer
  REDAUMExtractTimer aum_extract_timer_;
  //Timer to extract Discharge Extract messages
  REDDischargeExtractTimer discharge_extract_timer_;
  

  //NEIGHBORS MAINTENANCE
  /*The list of the node's neighbors (built by exploiting Hello packets)*/
  Neighbors_Table my_neighbors_table_;

  /*The table containing the last known neighbor*/
  int new_neighbor_;
    
  //REPLICA DISTRIBUTION        
  /*This represents the Replica Table.
    Each node stores the replicas it knows about and the nodes where they are stored*/
  Replica_Table  my_replica_table_;
  
  /*Each node stores the replicas it owns*/
  Replica_Table my_shared_replicas_;

  /*This variable contains information that a node should preserve from the reception of a delegation
   *packet to the send of the corresponding packet.*/
  Replication_Info* replication_forwarding_info_;

  /*This variable contains information that a node should preserve from the reception of a search
   *packet to the forwarding of the packet.*/
  Replication_Info* search_forwarding_info_;

  /*This variable contains information about the neighbors of the previous delegate*/
  Neighbors_Table* previous_delegate_neighbors_;
  
  /*The manager shared resource table*/
  Replica_Table  srt_;
  
  /*The manager message queue: it contains the messages to require new replicas when AUM check is done*/
  Message_Queue* request_new_replicas_queue_;

  /*The delegate discharge queue*/
  Discharge_Queue* discharge_queue_;
  
 
  

  //The next replica holder candidate
  int next_delegate_;
  /*The state of the current host.
    CODE:    0   free
             1   PROBED received
             2   location server*/ 
  int probed_state_;
  //I currently belong to the boundary, thus I reflect
  int reflect_;
  //The last neighbor reflected
  int im_after_reflection_;
    
 
  //The number of shared neighbors between the previous candidate acceptor and current node
  int shared_neigh_num_;
  int current_prober_;

  //Tables containing the number of shared neighbors of the previous candidate acceptor and current nodes neighbors
  Neighbors_Table my_probed_neighbors_;
  
  
  

  int probe_type_;
  
  //REPLICA SEARCH
  int current_replica_searcher_;
  int current_resource_searched_;
  int current_search_previous_neigh_[NODENUM];
  int current_search_previous_neigh_num_;
  int searched_probe_neighbors_num_;
  int search_hop_limit_;
  //The number of shared neighbors between the previous candidate acceptor and current node
  int search_shared_neigh_num_;
  int current_search_prober_;
  int found_;
  //The address of the last forwarder to avoid using routing protocols
  int search_fwd_;
  //A boolean to indicate whether a reply has already been sent to the searcher
  int replied_;
  //The replied replica placement
  int replied_replica_placement_;
  //Consider whether I have already been delegate for the search request or not.
  int yet_explored_for_search_;
  
};


/*=====================================================================================================
 *                                  REDMAN PARTNER
 *=====================================================================================================*/

/*
 *This REDMAN Partner helps filtering broadcast messages
 */

class RedmanPartner : public Connector {
public:
        RedmanPartner();
        void recv(Packet *p, Handler *h);

protected:
	      int command(int argc, const char*const*argv);
        ns_addr_t here_;
	      int redmanagent_;
        int mask_;
        int shift_;
};



#endif
