 /**
 *This file provides the interface of the REDMAN Packet Header.
 *These packets are the application layer packets used for RD and RR
 */

#ifndef __node_num__
#define __node_num__
const int NODENUM=60;
#endif

#ifndef __redman_packet__
#define __redman_packet__
#include <sys/types.h>
#include "common/packet.h"

/* =====================================================================
   Packet Formats...
   ===================================================================== */
#define REDMAN_DISC   	           0x01
#define REDMAN_HELLO  	           0x04
#define REDMAN_REPLICA    	       0x08

#define REDMAN_DISCHARGE           0x32

#define REDMAN_SEARCH     	       0x10
#define REDMAN_PROBE               0x12
#define REDMAN_PROBE_REP           0x14
#define REDMAN_CONFIRMATION        0x15
#define REDMAN_SEARCH_PROBE        0x16
#define REDMAN_SEARCH_PROBE_REP    0x18
#define REDMAN_SEARCH_REP          0x20
#define REDMAN_DEGREE_REQ          0x22
#define REDMAN_DEGREE_REP          0x24
#define REDMAN_NEW_REPLICA_PLACED  0x26
#define REDMAN_MANAGER_NOTIFY      0x28
#define REDMAN_REQUIRE_REPLICATION 0x30
#define REDMAN_AUM                 0x31





              /*
 * REDMAN Protocol Header Macros
 */
#define HDR_RED(p)		      ((struct hdr_red*)hdr_red::access(p))
#define HDR_RED_DISC(p)  	((struct hdr_red_disc*)hdr_red::access(p))
#define HDR_RED_HELLO(p)  	((struct hdr_red_hello*)hdr_red::access(p))
#define HDR_RED_REPLICA(p)	((struct hdr_red_replica*)hdr_red::access(p))
#define HDR_RED_SEARCH(p)	  ((struct hdr_red_search*)hdr_red::access(p))
#define HDR_RED_PROBE(p)	  ((struct hdr_red_probe*)hdr_red::access(p))
#define HDR_RED_PROBE_REP(p)	  ((struct hdr_red_probe_rep*)hdr_red::access(p))
#define HDR_RED_CONF(p)   ((struct hdr_red_conf*)hdr_red::access(p))
#define HDR_RED_SEARCH_PROBE(p)	  ((struct hdr_red_search_probe*)hdr_red::access(p))
#define HDR_RED_SEARCH_PROBE_REP(p)	  ((struct hdr_red_search_probe_rep*)hdr_red::access(p))
#define HDR_RED_SEARCH_REP(p)   ((struct hdr_red_search_rep*)hdr_red::access(p))
#define HDR_RED_DEGREE_REQ(p)   ((struct hdr_red_degree_req*)hdr_red::access(p))
#define HDR_RED_DEGREE_REP(p)   ((struct hdr_red_degree_rep*)hdr_red::access(p))
#define HDR_RED_NEW_REPLICA(p)	((struct hdr_red_new_replica*)hdr_red::access(p))
#define HDR_RED_MAN_NOT(p)      ((struct hdr_red_man_not*)hdr_red::access(p))
#define HDR_RED_REQ_REPL(p)	((struct hdr_red_req_repl*)hdr_red::access(p))
#define HDR_RED_AUM(p)          ((struct hdr_red_aum*)hdr_red::access(p))

/*
 *General REDMAN Header - shared by all REDMAN packets
 */
struct hdr_red {
        int           type;
        int           srcid;

      	// Header access methods
      	static int offset_; // required by PacketHeaderManager
      	inline static int& offset() { return offset_; }
      	inline static hdr_red* access(const Packet* p) {
      		return (hdr_red*) p->access(offset_);
      	}
};


/*
 *Boundary Identification protocol: discovery packet.
 *This packet is used by nodes to advertise their presence.
 */

struct hdr_red_disc {
        int          disc_type;
        int          disc_srcid;
};     

/*
 *REDMAN Topology control: hello packets.
 *These are used to determine the identity of nodes current neighbors.
 */            
struct hdr_red_hello {

        int         hello_type;
        int         hello_srcid;
        int         hello_part;

};

/*
 *REDMAN Replica Distribution: replica location packets.
 *These packets are used to determine where replicas should be placed.
 */
struct hdr_red_replica {

        int         replica_type;
        int         replica_srcid;
        int         replica_resource_delegate;
        int         replica_resource_id;
        int         replica_num_replicas_required;
        int         replica_acceptors[NODENUM];
        int         replica_acceptors_num;
        int         replica_hops_to_next;
        int         replica_previous_neigh[NODENUM];
        int         replica_previous_neigh_num;
        int         replica_reflection;
        //These permit to notify contemporaneously to the manager that the delegate that was carrying the
        //resource is exiting from the network
        int         replica_is_exiting;
        int         replica_exiting_delegate;
        //I will introduce explored

};

/*
 *REDMAN Replica Distribution: local neighbors probe packet.
 *This is used to find the neighbor sharing less neighbors with the previous node.
 */
struct hdr_red_probe {

        int         probe_type;
        int         probe_srcid;
        int         probe_previous_neigh[NODENUM];
        int         probe_previous_neigh_num;
        int         probe_replica_placement;
        int         probe_resource_id;

};

/*
 *REDMAN Replica Distribution: local neighbors probe reply packet.
 *Neighbors reply stating the number of neighbors they share with the previous node.
 */
struct hdr_red_probe_rep {

        int         probe_rep_type;
        int         probe_rep_srcid;
        int         probe_rep_num;
        int         probe_rep_found;
        int         probe_rep_replica_placement;
        int         probe_rep_state;

};

/*
 *REDMAN Replica Distribution: local neighbors confirmation of location service acceptance.
 */
struct hdr_red_conf {
  
        int         conf_type;
        int         conf_srcid;

};


/*
 *REDMAN Replica Retrieval: replica search packets.
 *These packets are used to find places where target resources are stored.
 */
struct hdr_red_search {

        int         search_type;
        int         search_srcid;
        int         search_searcher;
        int         search_resource_id;
        int         search_hop_limit;
        int         search_previous_neigh[NODENUM];
        int         search_previous_neigh_num;

};

/*
 *REDMAN Replica Retrieval: replica search reply packets.
 *These packets are used to reply to the searcher that a replica has/hasn't been found
 */
struct hdr_red_search_rep {

        int         search_rep_type;
        int         search_rep_srcid;
        int         search_rep_searcher;
        int         search_rep_resource_id;
        int         search_rep_replica_placement;

};

/*
 *REDMAN Replica Manager: replication degree request packets.
 *These packets are used by a delegate to require a replication degree for its resource(s)
 */
struct hdr_red_degree_req {
	
	int	    deg_req_type;
	int     deg_req_srcid;
  int	    deg_req_delegate;
	int	    deg_req_resource_id;
	int	    deg_req_sugg_degree;
	
};

/*
 *REDMAN Replica Manager: replication degree reply packets.
 *These packets are used by the manager to return the delegate the decided replication degree for its resource(s)
 */
struct hdr_red_degree_rep {
	
	int	    deg_rep_type;
	int     deg_rep_srcid;
	int	    deg_rep_resource_id;
	int	    deg_rep_degree;
	
};

/*
 *REDMAN Replica Manager: new copy placed packets.
 *These packets notify the manager that a new copy of the resource has been placed
 */
struct hdr_red_new_replica {
	
	int	    new_replica_type;
	int     new_replica_srcid;
  int	    new_replica_delegate;
	int	    new_replica_resource_id;
  int     new_replica_is_delegate_exiting;
  int     new_replica_exiting_delegate;
	
};

/*
 *REDMAN Replica Manager: the current node is out of the network
 *These packets notify the manager that the delegate has run out of the network
 */
struct hdr_red_man_not {
	
	int	    man_not_type;
	int     man_not_srcid;
  int	    man_not_delegate;
	int	    man_not_resource_id[NODENUM];
	int 	  man_not_resource_id_num;
	
};

/*
 *REDMAN Replica Manager: the manager requires a new replication of the marked resource
 *These packets are used by the manager to require a new replication from a node hosting a replica
 */
struct hdr_red_req_repl {
	
	int	    req_repl_type;
	int     req_repl_srcid;
  int	    req_repl_resource_id; //[NODENUM]
	int     req_repl_copies;
	//int         req_repl_resource_id_num;
	
};

/*
 *REDMAN Replica Manager: the delegate complies to AUM protocol
 *These packets are used by delegates to periodically confirm their presence and the resources they own
 */
struct hdr_red_aum {
	
	int	    aum_type;
	int     aum_srcid;
	int     aum_delegate;
  int	    aum_resource_id[NODENUM];
	int     aum_resource_id_num;
	
};

#endif
